# Copyright (c) 2016 Juniper Networks Inc.
# All Rights Reserved.
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

BGPVPN = 'bgpvpn'
BGPVPNS = f'{BGPVPN}s'

NETWORK_RESOURCE_NAME = 'network'
NETWORK_ASSOCIATION = f'{NETWORK_RESOURCE_NAME}_association'
NETWORK_ASSOCIATIONS = f'{NETWORK_ASSOCIATION}s'

ROUTER_RESOURCE_NAME = 'router'
ROUTER_ASSOCIATION = f'{ROUTER_RESOURCE_NAME}_association'
ROUTER_ASSOCIATIONS = f'{ROUTER_ASSOCIATION}s'

PORT_RESOURCE_NAME = 'port'
PORT_ASSOCIATION = f'{PORT_RESOURCE_NAME}_association'
PORT_ASSOCIATIONS = f'{PORT_ASSOCIATION}s'
